package be.iminds.ilabt.jfed.lowlevel.api;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;

/**
 * StitchingComputationService
 * <p/>
 * see https://dragon.maxgigapop.net/twiki/bin/view/GENI/NetworkStitchingAPI
 */
public class StitchingComputationService extends AbstractApi {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(StitchingComputationService.class);

    public StitchingComputationService(@Nonnull Logger logger, @Nonnull JFedPreferences jFedPreferences) {
        this(logger, jFedPreferences.getRetrySettings(), jFedPreferences);
    }

    public StitchingComputationService(@Nonnull Logger logger, RetrySettings retrySettings, @Nonnull JFedPreferences jFedPreferences) {
        super(logger, retrySettings, new ApiInfo.Api(ApiInfo.ApiName.GENI_SCS, 1), jFedPreferences);
    }

    /**
     * A human readable name for the implemented API
     *
     * @return "Stitching Computation Service";
     */
    static public String getApiName() {
        return "Stitching Computation Service";
    }

    @Override
    protected boolean isBusyReply(be.iminds.ilabt.jfed.lowlevel.XMLRPCCallDetails res) {
        if (res instanceof XMLRPCCallDetailsWithCodeValueError) {
            XMLRPCCallDetailsWithCodeValueError geniDetails = (XMLRPCCallDetailsWithCodeValueError) res;
            Hashtable codeStruct = (Hashtable) geniDetails.getResultCode();
            int code = (Integer) codeStruct.get("geni_code");
            return GeniAMResponseCode.getByCode(code).isBusy();
        }
        return false;
    }

    /**
     * A human readable name for the implemented API
     *
     * @return "Stitching Computation Service";
     */
    @Override
    public String getName() {
        return getApiName();
    }

    @ApiMethod(order = 1, hint = "GetVersion call: Get static version and configuration information about this SCS. There is no documentation about this call, but it does exist.")
    public SCSReply<Hashtable> getVersion(SfaConnection con) throws JFedException {
        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "GetVersion", new Vector(), null);
        Object val = res.getResultValueObject();
        SCSReply<Hashtable> r = new SCSReply<Hashtable>(res);
        log(res, r, "getVersion", "GetVersion", con, null);
        return r;
    }

    @ApiMethod(order = 2, hint = "ComputePath call: Compute a Network Stitching Path. See https://wiki.maxgigapop.net/twiki/bin/view/GENI/NetworkStitchingAPI")
    public SCSReply<ComputePathResult> computePath(SfaConnection con,
                                                   @ApiMethodParameter(name = "sliceUrn", hint = "The slice on which everything needs to be created", parameterType = ApiMethodParameterType.SLICE_URN)
                                                   String sliceUrn,
                                                   @ApiMethodParameter(name = "requestRspec", hint = "request rspec with or without stitching extensions", parameterType = ApiMethodParameterType.RSPEC_STRING)
                                                   String requestRspec,
                                                   @ApiMethodParameter(name = "requestOptions", required = false, hint = "optional options")
                                                   Hashtable requestOptions) throws JFedException {
        //this is documented
//        Vector args = new Vector(3);
//        args.add(sliceUrn);
//        args.add(requestRspec);
//        if (requestOptions == null)
//            args.add(new Hashtable());
//        else
//            args.add(requestOptions);

        //but this is what it really is
        Vector args = new Vector(1);
        Hashtable argsTable = new Hashtable();
        args.add(argsTable);
        argsTable.put("slice_urn", sliceUrn);
        argsTable.put("request_rspec", requestRspec);
        if (requestOptions == null)
            argsTable.put("request_options", new Hashtable());
        else
            argsTable.put("request_options", requestOptions);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "ComputePath", args, null);
        SCSReply<ComputePathResult> r = null;
        if (res.getResult() != null) {
            if (res.getResultValueObject() != null && res.getResultValueObject() instanceof Hashtable) {
                try {
                    Hashtable resultHashtable = (Hashtable) res.getResultValueObject();

                    String rspec = null;
                    Object serviceRspecObject = resultHashtable.get("service_rspec");
                    if (serviceRspecObject != null) {
                        if (serviceRspecObject instanceof String)
                            rspec = (String) serviceRspecObject;
                        else
                            LOG.warn("WARNING: SCS ComputePath service_rspec is of type " + serviceRspecObject.getClass().getName() + "  value=\"" + serviceRspecObject + "\"");
                    }

                    Hashtable workFlowData = null;
                    Object workFlowDataObject = resultHashtable.get("workflow_data");
                    if (workFlowDataObject != null && workFlowDataObject instanceof Hashtable)
                        workFlowData = (Hashtable) workFlowDataObject;

                    r = new SCSReply<ComputePathResult>(res, new ComputePathResult(workFlowData, rspec));
                } catch (Exception e) {
                    LOG.error("Exception parsing SCS ComputePath response: " + e.getMessage(), e);
                }
            }
//            else if (res.getResultOutput() != null) {
//                throw new JFedException("Received an error from SCS: " + res.getResultOutput(), res, null);
//            }
        }
        if (r == null)
            r = new SCSReply<ComputePathResult>(res, null);
        log(res, r, "ComputePath", "ComputePath", con, null);
        return r;
    }



    @ApiMethod(order = 2, hint = "ComputePath call with extra options: Compute a Network Stitching Path. See https://wiki.maxgigapop.net/twiki/bin/view/GENI/NetworkStitchingAPI")
    public SCSReply<ComputePathResult> computePathWithOptions(SfaConnection con,
                                                   @ApiMethodParameter(name = "sliceUrn", hint = "The slice on which everything needs to be created", parameterType = ApiMethodParameterType.SLICE_URN)
                                                   String sliceUrn,
                                                   @ApiMethodParameter(name = "requestRspec", hint = "request rspec with or without stitching extensions", parameterType = ApiMethodParameterType.RSPEC_STRING)
                                                   String requestRspec,
                                                   @ApiMethodParameter(name = "requestOptions", required = false, hint = "optional options")
                                                   Hashtable requestOptions,

                                                   //known requestOptions
                                                   @ApiMethodParameter(name = "geniHoldPath", required = false, guiDefault = "false", hint = "optional, default=false. SCS to hold the path exclusively for a preconfigured holding period to avoid contention")
                                                   Boolean geniHoldPath,
                                                   @ApiMethodParameter(name = "geniStartTime", required = false, hint = "optional, default=now, projected starting time for the stitching path(s)")
                                                   Date geniStartTime,
                                                   @ApiMethodParameter(name = "geniEndTime", required = false, hint = "optional, default=infinite, projected expiration time for the stitching path(s)")
                                                   Date geniEndTime,
                                                   @ApiMethodParameter(name = "geniRoutingProfile", required = false, guiDefault = "{ 'hop_inclusion_list': [] , 'hop_exclusion_list': [] }", hint = "optional,  a struct of {path_id: routing_profile} map. The path_id corresponding to a stitching path id in request_rspec. routing_profile is a struct of two optional items {'hop_inclusion_list': array of urn strings} and {'hop_exclusion_list': array of urn strings}.")
                                                   Hashtable geniRoutingProfile,
                                                   @ApiMethodParameter(name = "geniWorkflowData", required = false, guiDefault = "true", hint = "optional, default=true, SCS to return workflow data in support of user / agent / tool workflow logic.")
                                                   Boolean geniWorkflowData

                                                   ) throws JFedException {
        if (requestOptions == null)
            requestOptions = new Hashtable();

        if (geniHoldPath != null)
            requestOptions.put("geni_hold_path", geniHoldPath.booleanValue());
        if (geniStartTime != null)
            requestOptions.put("geni_start_time", RFC3339Util.dateToRFC3339String(geniStartTime, true/*forceZuluhack*/));
        if (geniEndTime != null)
            requestOptions.put("geni_end_time", RFC3339Util.dateToRFC3339String(geniEndTime, true/*forceZuluhack*/));
        if (geniRoutingProfile != null)
            requestOptions.put("geni_routing_profile", geniRoutingProfile);
        if (geniWorkflowData != null)
            requestOptions.put("geni_workflow_data", geniWorkflowData.booleanValue());

        return computePath(con, sliceUrn, requestRspec, requestOptions);
    }

    public static class SCSReply<T> implements SfaApiCallReply<T> {
        private final XMLRPCCallDetailsWithCodeValueError xmlRpcCallDetailsWithCodeValueError;
        private final SCSResponseCode genicode;
        private final T val;
        /* output is typically set only on error*/
        private final String output;
        private final Hashtable rawResult;

        public SCSReply(XMLRPCCallDetailsWithCodeValueError res) {
            this.xmlRpcCallDetailsWithCodeValueError = res;
            this.rawResult = res.getResult();
            Hashtable code = (Hashtable) res.getResultCode();
            int intCode = (Integer) code.get("geni_code");
            this.genicode = SCSResponseCode.getByCode(intCode);
            T v;
            try {
                v = (T) res.getResultValueObject();
            } catch (Exception e) {
                v = null;
            }
            this.val = v;
            this.output = res.getResultOutput();
        }

        public SCSReply(XMLRPCCallDetailsWithCodeValueError res, T val) {
            this.xmlRpcCallDetailsWithCodeValueError = res;
            this.rawResult = res.getResult();
            Hashtable code = (Hashtable) res.getResultCode();
            int intCode = (Integer) code.get("geni_code");
            this.genicode = SCSResponseCode.getByCode(intCode);
            this.val = val;
            this.output = res.getResultOutput();
        }

        public int getCode() {
            return genicode.getCode();
        }

        public SCSResponseCode getGeniResponseCode() {
            return genicode;
        }

        public T getValue() {
            return val;
        }

        public String getOutput() {
            return output;
        }

        public XMLRPCCallDetailsWithCodeValueError getXmlRpcCallDetailsWithCodeValueError() {
            return xmlRpcCallDetailsWithCodeValueError;
        }

        @Override
        public Hashtable getRawResult() {
            return rawResult;
        }

        @Override
        public XMLRPCCallDetailsWithCodeValueError getXMLRPCCallDetailsWithCodeValueError() {
            return xmlRpcCallDetailsWithCodeValueError;
        }

        @Override
        public XMLRPCCallDetails getXMLRPCCallDetails() {
            return xmlRpcCallDetailsWithCodeValueError;
        }
    }

    public static class ComputePathResult {
        private final String serviceRspec;
        private final Hashtable workflowData;

        public ComputePathResult(Hashtable workflowData, String serviceRspec) {
            this.workflowData = workflowData;
            this.serviceRspec = serviceRspec;
        }

        public String getServiceRspec() {
            return serviceRspec;
        }

        public Hashtable getWorkflowData() {
            return workflowData;
        }

        @Override
        public String toString() {
            return "ComputePathResult{" +
                    "serviceRspec='" + serviceRspec + '\'' +
                    ", workflowData=" + workflowData +
                    '}';
        }
    }
}
